"use strict";
var config = require("./config").config;
var EventTarget = require("./events").EventTarget;
var CancellationError = require("./cancellation_error").CancellationError;

function objectOrFunction(x) {
  return isFunction(x) || (typeof x === "object" && x !== null);
}

function isFunction(x){
  return typeof x === "function";
}

var Promise = function(resolver, canceller) {
  var promise = this,
  resolved = false;

  if (typeof resolver !== 'function') {
    throw new TypeError('You must pass a resolver function as the sole argument to the promise constructor');
  }

  if ((canceller !== undefined) && (typeof canceller !== 'function')) {
    throw new TypeError('You can only pass a canceller function' +
      ' as the second argument to the promise constructor');
  }

  if (!(promise instanceof Promise)) {
    return new Promise(resolver, canceller);
  }

  var resolvePromise = function(value) {
    if (resolved) { return; }
    resolved = true;
    resolve(promise, value);
  };

  var rejectPromise = function(value) {
    if (resolved) { return; }
    resolved = true;
    reject(promise, value);
  };

  var notifyPromise = function(value) {
    if (resolved) { return; }
    notify(promise, value);
  };

  this.on('promise:failed', function(event) {
    this.trigger('error', { detail: event.detail });
  }, this);

  this.on('error', onerror);

  this.cancel = function () {
    // For now, simply reject the promise and does not propagate the cancel
    // to parent or children
    if (resolved) { return; }
    if (canceller !== undefined) {
      try {
        canceller();
      } catch (e) {
        rejectPromise(e);
        return;
      }
    }
    // Trigger cancel?
    rejectPromise(new CancellationError());
  };

  try {
    resolver(resolvePromise, rejectPromise, notifyPromise);
  } catch(e) {
    rejectPromise(e);
  }
};

function onerror(event) {
  if (config.onerror) {
    config.onerror(event.detail);
  }
}

var invokeCallback = function(type, promise, callback, event) {
  var hasCallback = isFunction(callback),
      value, error, succeeded, failed;

  if (promise.isFulfilled) { return; }
  if (promise.isRejected) { return; }

  if (hasCallback) {
    try {
      value = callback(event.detail);
      succeeded = true;
    } catch(e) {
      failed = true;
      error = e;
    }
  } else {
    value = event.detail;
    succeeded = true;
  }

  if (handleThenable(promise, value)) {
    return;
  } else if (hasCallback && succeeded) {
    resolve(promise, value);
  } else if (failed) {
    reject(promise, error);
  } else if (type === 'resolve') {
    resolve(promise, value);
  } else if (type === 'reject') {
    reject(promise, value);
  }
};


var invokeNotifyCallback = function(promise, callback, event) {
  var value;
  if (typeof callback === 'function') {
    try {
      value = callback(event.detail);
    } catch (e) {
      // stop propagating
      return;
    }
    notify(promise, value);
  } else {
    notify(promise, event.detail);
  }
};

Promise.prototype = {
  constructor: Promise,

  isRejected: undefined,
  isFulfilled: undefined,
  rejectedReason: undefined,
  fulfillmentValue: undefined,

  then: function(done, fail, progress) {
    this.off('error', onerror);

    var thenPromise = new this.constructor(function() {},
        function () {
          thenPromise.trigger('promise:cancelled', {});
        });

    if (this.isFulfilled) {
      config.async(function(promise) {
        invokeCallback('resolve', thenPromise, done, { detail: promise.fulfillmentValue });
      }, this);
    }

    if (this.isRejected) {
      config.async(function(promise) {
        invokeCallback('reject', thenPromise, fail, { detail: promise.rejectedReason });
      }, this);
    }

    this.on('promise:resolved', function(event) {
      invokeCallback('resolve', thenPromise, done, event);
    });

    this.on('promise:failed', function(event) {
      invokeCallback('reject', thenPromise, fail, event);
    });

    this.on('promise:notified', function (event) {
      invokeNotifyCallback(thenPromise, progress, event);
    });

    return thenPromise;
  },

  fail: function(fail) {
    return this.then(null, fail);
  },

  always: function(fail) {
    return this.then(fail, fail);
  }
};

EventTarget.mixin(Promise.prototype);

function resolve(promise, value) {
  if (promise === value) {
    fulfill(promise, value);
  } else if (!handleThenable(promise, value)) {
    fulfill(promise, value);
  }
}

function handleThenable(promise, value) {
  var then = null,
  resolved;

  try {
    if (promise === value) {
      throw new TypeError("A promises callback cannot return that same promise.");
    }

    if (objectOrFunction(value)) {
      then = value.then;

      if (isFunction(then)) {
        if (isFunction(value.on)) {
          value.on('promise:notified', function (event) {
            notify(promise, event.detail);
          });
        }
        promise.on('promise:cancelled', function(event) {
          if (isFunction(value.cancel)) {
            value.cancel();
          }
        });
        then.call(value, function(val) {
          if (resolved) { return true; }
          resolved = true;

          if (value !== val) {
            resolve(promise, val);
          } else {
            fulfill(promise, val);
          }
        }, function(val) {
          if (resolved) { return true; }
          resolved = true;

          reject(promise, val);
        });

        return true;
      }
    }
  } catch (error) {
    reject(promise, error);
    return true;
  }

  return false;
}

function fulfill(promise, value) {
  config.async(function() {
    if (promise.isFulfilled) { return; }
    if (promise.isRejected) { return; }
    promise.trigger('promise:resolved', { detail: value });
    promise.isFulfilled = true;
    promise.fulfillmentValue = value;
  });
}

function reject(promise, value) {
  config.async(function() {
    if (promise.isFulfilled) { return; }
    if (promise.isRejected) { return; }
    promise.trigger('promise:failed', { detail: value });
    promise.isRejected = true;
    promise.rejectedReason = value;
  });
}

function notify(promise, value) {
  config.async(function() {
    promise.trigger('promise:notified', { detail: value });
  });
}


exports.Promise = Promise;