/*global describe, specify, it, assert */

var local = (typeof global === "undefined") ? this : global,
    oldSetTimeout, newSetTimeout;
local.setTimeout = local.setTimeout;
oldSetTimeout = local.setTimeout;
newSetTimeout = function(callback) {
  var errorWasThrown;
  try {
    callback.call(this, arguments);
  } catch(e) {
    errorWasThrown = true;
  }
};

if (typeof Object.getPrototypeOf !== "function") {
  Object.getPrototypeOf = "".__proto__ === String.prototype
    ? function (object) {
      return object.__proto__;
    }
    : function (object) {
      // May break if the constructor has been tampered with
      return object.constructor.prototype;
    };
}

function objectEquals(obj1, obj2) {
  for (var i in obj1) {
    if (obj1.hasOwnProperty(i)) {
      if (!obj2.hasOwnProperty(i)) return false;
      if (obj1[i] != obj2[i]) return false;
    }
  }
  for (var i in obj2) {
    if (obj2.hasOwnProperty(i)) {
      if (!obj1.hasOwnProperty(i)) return false;
      if (obj1[i] != obj2[i]) return false;
    }
  }
  return true;
}

describe("RSVP extensions", function() {
  describe("self fulfillment", function(){
    it("treats self fulfillment as the recursive base case", function(done){
      var aDefer = new RSVP.defer(),
      bDefer = new RSVP.defer(),
      promiseA = aDefer.promise,
      promiseB = bDefer.promise;

      promiseA.then(function(a){
        setTimeout(function(){
          bDefer.resolve(promiseB);
        }, 1);

        return promiseB;
      });

      promiseB.then(function(c){
        done();
      })

      aDefer.resolve(promiseA);
    });
  });

  describe("Promise constructor", function() {
    it('should exist and have length 2', function() {
      assert(RSVP.Promise);
      assert.equal(RSVP.Promise.length, 2);
    });

    it('should fulfill if `resolve` is called with a value', function(done) {
      var promise = new RSVP.Promise(function(resolve) { resolve('value'); });

      promise.then(function(value) {
        assert.equal(value, 'value');
        done();
      });
    });

    it('should reject if `reject` is called with a reason', function(done) {
      var promise = new RSVP.Promise(function(resolve, reject) { reject('reason'); });

      promise.then(function() {
        assert(false);
        done();
      }, function(reason) {
        assert.equal(reason, 'reason');
        done();
      });
    });

    it('should be a constructor', function() {
      var promise = new RSVP.Promise(function() {});

      assert.equal(Object.getPrototypeOf(promise), RSVP.Promise.prototype, '[[Prototype]] equals Promise.prototype');
      assert.equal(promise.constructor, RSVP.Promise, 'constructor property of instances is set correctly');
      assert.equal(RSVP.Promise.prototype.constructor, RSVP.Promise, 'constructor property of prototype is set correctly');
    });

    it('should work without `new`', function(done) {
      var promise = RSVP.Promise(function(resolve) { resolve('value'); });

      promise.then(function(value) {
        assert.equal(value, 'value');
        done();
      });
    });

    it('should throw a `TypeError` if not given a function', function() {
      assert.throws(function () {
        var promise = new RSVP.Promise();
      }, TypeError);

      assert.throws(function () {
        var promise = new RSVP.Promise({});
      }, TypeError);

      assert.throws(function () {
        var promise = new RSVP.Promise('boo!');
      }, TypeError);
    });

    it('should throw a `TypeError` if the `canceller` is provided ' +
       'and not a function', function () {
      assert.throws(function () {
        var promise = new RSVP.Promise(function () {}, {});
      }, TypeError);

      assert.throws(function () {
        var promise = new RSVP.Promise(function () {}, 'boo!');
      }, TypeError);
    });

    it('should reject on resolver exception', function(done) {
      var promise = RSVP.Promise(function() {
        throw 'error';
      }).then(null, function(e) {
        assert.equal(e, 'error');
        done();
      });
    });

    it('should not resolve multiple times', function(done) {
      var resolver, rejector, fulfilled = 0, rejected = 0;
      var thenable = {
        then: function(resolve, reject) {
          resolver = resolve;
          rejector = reject;
        }
      };

      var promise = RSVP.Promise(function(resolve) {
        resolve(1);
      });

      promise.then(function(value){
        return thenable;
      }).then(function(value){
        fulfilled++;
      }, function(reason) {
        rejected++;
      });

      setTimeout(function() {
        resolver(1);
        resolver(1);
        rejector(1);
        rejector(1);

        setTimeout(function() {
          assert.equal(fulfilled, 1);
          assert.equal(rejected, 0);
          done();
        }, 20);
      }, 20);

    });

    it('notify should exist', function (done) {
      var exists, promise = new RSVP.Promise(function (resolve, reject, notify) {
        exists = typeof notify === 'function';
      });
      setTimeout(function () {
        assert(exists);
        done();
      }, 20);
    });

    it('should notify if `notify` is called with a value', function (done) {
      var promise = new RSVP.Promise(function (resolve, reject, notify) {
        notify('foo');
      });
      promise.then(null, null, function (foo) {
        assert.equal(foo, 'foo');
        done();
      });
    });

    it('should notify twice before resolve', function (done) {
      var notified = 0, promise = new RSVP.Promise(function (resolve, reject, notify) {
        notify('foo');
        notify('bar');
        resolve('end');
        notify('should not be notified');
      });
      promise.then(function () {
        var currently_notified = notified;
        setTimeout(function () {
          assert.equal(currently_notified, 2);
          assert.equal(notified, 2);
          done();
        }, 50);
      }, null, function () {
        notified += 1;
      });
    });

    it('notify should propagate on two then', function (done) {
      var last_notified = false, promise = new RSVP.Promise(function (resolve, reject, notify) {
        notify('foo');
      });
      promise.then(null, null, function (foo) {
        assert.equal(foo, 'foo');
        return 'bar';
      }).then(null, null, function (bar) {
        assert.equal(bar, 'bar');
        setTimeout(function () {
          assert(!last_notified);
          done();
        }, 50);
        throw 'baz';
      }).then(null, null, function () {
        last_notified = true;
      });
    });

    it('notify should propagate to parent `then`', function (done) {
      var i = 0, responses = [];

      new RSVP.Promise(function (resolve) {
        resolve();
      }).then(function () {
        return new RSVP.Promise(function (resolve, reject, notify) {
          notify(1);
          notify(2);
          resolve(3);
        });
      }).then(function (answer) {
        responses.push(answer);
      }, null, function (notification) {
        responses.push(notification);
      });

      setTimeout(function () {
        assert(responses[0] === 1, "Notified once");
        assert(responses[1] === 2, "Notified twice");
        assert(responses[2] === 3, "Fulfilled after notifications");
        done();
      }, 50);
    });

    describe('assimilation', function() {
      it('should assimilate if `resolve` is called with a fulfilled promise', function(done) {
        var originalPromise = new RSVP.Promise(function(resolve) { resolve('original value'); });
        var promise = new RSVP.Promise(function(resolve) { resolve(originalPromise); });

        promise.then(function(value) {
          assert.equal(value, 'original value');
          done();
        });
      });

      it('should assimilate if `resolve` is called with a rejected promise', function(done) {
        var originalPromise = new RSVP.Promise(function(resolve, reject) { reject('original reason'); });
        var promise = new RSVP.Promise(function(resolve) { resolve(originalPromise); });

        promise.then(function() {
          assert(false);
          done();
        }, function(reason) {
          assert.equal(reason, 'original reason');
          done();
        });
      });

      it('should assimilate if `resolve` is called with a fulfilled thenable', function(done) {
        var originalThenable = {
          then: function (onFulfilled) {
            setTimeout(function() { onFulfilled('original value'); }, 0);
          }
        };
        var promise = new RSVP.Promise(function(resolve) { resolve(originalThenable); });

        promise.then(function(value) {
          assert.equal(value, 'original value');
          done();
        });
      });

      it('should assimilate if `resolve` is called with a rejected thenable', function(done) {
        var originalThenable = {
          then: function (onFulfilled, onRejected) {
            setTimeout(function() { onRejected('original reason'); }, 0);
          }
        };
        var promise = new RSVP.Promise(function(resolve) { resolve(originalThenable); });

        promise.then(function() {
          assert(false);
          done();
        }, function(reason) {
          assert.equal(reason, 'original reason');
          done();
        });
      });


      it('should assimilate two levels deep, for fulfillment of self fulfilling promises', function(done) {
        var originalPromise, promise;
        originalPromise = new RSVP.Promise(function(resolve) {
          setTimeout(function() {
            resolve(originalPromise);
          }, 0)
        });

        promise = new RSVP.Promise(function(resolve) {
          setTimeout(function() {
            resolve(originalPromise);
          }, 0);
        });

        promise.then(function(value) {
          assert.equal(value, originalPromise);
          done();
        });
      });

      it('should assimilate two levels deep, for fulfillment', function(done) {
        var originalPromise = new RSVP.Promise(function(resolve) { resolve('original value'); });
        var nextPromise = new RSVP.Promise(function(resolve) { resolve(originalPromise); });
        var promise = new RSVP.Promise(function(resolve) { resolve(nextPromise); });

        promise.then(function(value) {
          assert.equal(value, 'original value');
          done();
        });
      });

      it('should assimilate two levels deep, for rejection', function(done) {
        var originalPromise = new RSVP.Promise(function(resolve, reject) { reject('original reason'); });
        var nextPromise = new RSVP.Promise(function(resolve) { resolve(originalPromise); });
        var promise = new RSVP.Promise(function(resolve) { resolve(nextPromise); });

        promise.then(function() {
          assert(false);
          done();
        }, function(reason) {
          assert.equal(reason, 'original reason');
          done();
        });
      });

      it('should assimilate three levels deep, mixing thenables and promises (fulfilled case)', function(done) {
        var originalPromise = new RSVP.Promise(function(resolve) { resolve('original value'); });
        var intermediateThenable = {
          then: function (onFulfilled) {
            setTimeout(function() { onFulfilled(originalPromise); }, 0);
          }
        };
        var promise = new RSVP.Promise(function(resolve) { resolve(intermediateThenable); });

        promise.then(function(value) {
          assert.equal(value, 'original value');
          done();
        });
      });

      it('should assimilate three levels deep, mixing thenables and promises (rejected case)', function(done) {
        var originalPromise = new RSVP.Promise(function(resolve, reject) { reject('original reason'); });
        var intermediateThenable = {
          then: function (onFulfilled) {
            setTimeout(function() { onFulfilled(originalPromise); }, 0);
          }
        };
        var promise = new RSVP.Promise(function(resolve) { resolve(intermediateThenable); });

        promise.then(function() {
          assert(false);
          done();
        }, function(reason) {
          assert.equal(reason, 'original reason');
          done();
        });
      });
    });
  });

  describe("RSVP.defer", function() {
    specify("It should return a resolver and promise together", function(done) {
      var deferred = RSVP.defer(), value = {};

      // resolve first to confirm that the semantics are async
      deferred.resolve(value);

      deferred.promise.then(function(passedValue) {
        assert(passedValue === value);
        done();
      });
    });

    specify("The provided resolver should support rejection", function(done) {
      var deferred = RSVP.defer(), reason = {};

      // resolve first to confirm that the semantics are async
      deferred.reject(reason);

      deferred.promise.then(null, function(passedReason) {
        assert(passedReason === reason);
        done();
      });
    });
  });

  describe("RSVP.denodeify", function() {
    specify('it should exist', function() {
      assert(RSVP.denodeify);
    });

    specify('calls node function with any arguments passed', function(done) {
      var args = null;

      function nodeFunc(arg1, arg2, arg3, cb) {
        args = [arg1, arg2, arg3];
        cb();
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);

      denodeifiedFunc(1, 2, 3).then(function() {
        assert(objectEquals(args, [1, 2, 3]));
        done();
      });
    });

    specify('calls node function with same thisArg', function(done) {
      var thisArg = null;

      function nodeFunc(cb) {
        thisArg = this;
        cb();
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);
      var expectedThis = { expect: "me" };

      denodeifiedFunc.call(expectedThis).then(function() {
        assert.equal(thisArg, expectedThis);
        done();
      });
    });

    specify('waits for promise/thenable arguments to settle before passing them to the node function', function(done) {
      var args = null;

      function nodeFunc(arg1, arg2, arg3, cb) {
        args = [arg1, arg2, arg3];
        cb();
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);

      var promise = new RSVP.Promise(function(resolve) { resolve(1); });
      var thenable = { then: function (onFulfilled) { onFulfilled(2); } };
      var nonPromise = 3;
      denodeifiedFunc(promise, thenable, nonPromise).then(function() {
        assert(objectEquals(args, [1, 2, 3]));
        done();
      });
    });

    specify('fulfilled with value if node function calls back with a single argument', function(done) {
      function nodeFunc(cb) {
        cb(null, 'nodeFuncResult');
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);

      denodeifiedFunc().then(function(value) {
        assert.equal(value, 'nodeFuncResult');
        done();
      });
    });

    specify('fulfilled with array if node function calls back with multiple arguments', function(done) {
      function nodeFunc(cb) {
        cb(null, 1, 2, 3);
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);

      denodeifiedFunc().then(function(value) {
        assert(objectEquals(value, [1, 2, 3]));
        done();
      });
    });

    specify('rejected if node function calls back with error', function(done) {
      function nodeFunc(cb) {
        cb('bad!');
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);

      denodeifiedFunc().then(function() {
        assert(false);
        done();
      }, function(reason) {
        assert.equal(reason, 'bad!');
        done();
      });
    });

    specify('rejected if node function throws an exception synchronously', function(done) {
      function nodeFunc(cb) {
        throw 'bad!';
      }

      var denodeifiedFunc = RSVP.denodeify(nodeFunc);

      denodeifiedFunc().then(function() {
        assert(false);
        done();
      }, function(reason) {
        assert.equal(reason, 'bad!');
        done();
      });
    });

    specify('integration test showing how awesome this can be', function(done) {
      function readFile(fileName, cb) {
        setTimeout(function() {
          cb(null, 'contents of ' + fileName);
        }, 0);
      }

      var writtenTo = null;
      function writeFile(fileName, text, cb) {
        setTimeout(function () {
          writtenTo = [fileName, text];
          cb();
        }, 0);
      }

      var denodeifiedReadFile = RSVP.denodeify(readFile);
      var denodeifiedWriteFile = RSVP.denodeify(writeFile);

      denodeifiedWriteFile('dest.txt', denodeifiedReadFile('src.txt')).then(function () {
        assert(objectEquals(writtenTo, ['dest.txt', 'contents of src.txt']));
        done();
      });
    });
  });

  describe("RSVP.hash", function() {
    it('should exist', function() {
      assert(RSVP.hash);
    });

    specify('fulfilled only after all of the promise values are fulfilled', function(done) {
      var firstResolved, secondResolved, firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve) {
        firstResolver = resolve;
      });
      first.then(function() {
        firstResolved = true;
      });

      var second = new RSVP.Promise(function(resolve) {
        secondResolver = resolve;
      });
      second.then(function() {
        secondResolved = true;
      });

      setTimeout(function() {
        firstResolver(true);
      }, 0);

      setTimeout(function() {
        secondResolver(true);
      }, 0);

      RSVP.hash({ first: first, second: second }).then(function(values) {
        assert(values.first);
        assert(values.second);
        done();
      });
    });

    specify('rejected as soon as a promise is rejected', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      RSVP.hash({ first: first, second: second }).then(function() {
        assert(false);
      }, function() {
        assert(first.isRejected);
        assert(!second.isResolved);
        done();
      });
    });

    specify('resolves an empty hash passed to RSVP.all()', function(done) {
      RSVP.hash({}).then(function(results) {
        assert(objectEquals(results, {}), 'expected fulfillment');
        done();
      });
    });

    specify('works with a mix of promises and thenables and non-promises', function(done) {
      var promise = new RSVP.Promise(function(resolve) { resolve(1); });
      var syncThenable = { then: function (onFulfilled) { onFulfilled(2); } };
      var asyncThenable = { then: function (onFulfilled) { setTimeout(function() { onFulfilled(3); }, 0); } };
      var nonPromise = 4;

      RSVP.hash({ promise: promise, syncThenable: syncThenable, asyncThenable: asyncThenable, nonPromise: nonPromise }).then(function(results) {
        assert(objectEquals(results, { promise: 1, syncThenable: 2, asyncThenable: 3, nonPromise: 4 }));
        done();
      });
    });

  });

  describe("RSVP.all", function() {
    it('should exist', function() {
      assert(RSVP.all);
    });

    it('throws when not passed an array', function() {
      assert.throws(function () {
        var all = RSVP.all();
      }, TypeError);

      assert.throws(function () {
        var all = RSVP.all('');
      }, TypeError);

      assert.throws(function () {
        var all = RSVP.all({});
      }, TypeError);
    });

    specify('fulfilled only after all of the other promises are fulfilled', function(done) {
      var firstResolved, secondResolved, firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve) {
        firstResolver = resolve;
      });
      first.then(function() {
        firstResolved = true;
      });

      var second = new RSVP.Promise(function(resolve) {
        secondResolver = resolve;
      });
      second.then(function() {
        secondResolved = true;
      });

      setTimeout(function() {
        firstResolver(true);
      }, 0);

      setTimeout(function() {
        secondResolver(true);
      }, 0);

      RSVP.all([first, second]).then(function() {
        assert(firstResolved);
        assert(secondResolved);
        done();
      });
    });

    specify('rejected as soon as a promise is rejected', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      RSVP.all([first, second]).then(function() {
        assert(false);
      }, function() {
        assert(first.isRejected);
        assert(!second.isResolved);
        done();
      });
    });

    specify('passes the resolved values of each promise to the callback in the correct order', function(done) {
      var firstResolver, secondResolver, thirdResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      var third = new RSVP.Promise(function(resolve, reject) {
        thirdResolver = { resolve: resolve, reject: reject };
      });

      thirdResolver.resolve(3);
      firstResolver.resolve(1);
      secondResolver.resolve(2);

      RSVP.all([first, second, third]).then(function(results) {
        assert(results.length === 3);
        assert(results[0] === 1);
        assert(results[1] === 2);
        assert(results[2] === 3);
        done();
      });
    });

    specify('resolves an empty array passed to RSVP.all()', function(done) {
      RSVP.all([]).then(function(results) {
        assert(results.length === 0);
        done();
      });
    });

    specify('works with a mix of promises and thenables and non-promises', function(done) {
      var promise = new RSVP.Promise(function(resolve) { resolve(1); });
      var syncThenable = { then: function (onFulfilled) { onFulfilled(2); } };
      var asyncThenable = { then: function (onFulfilled) { setTimeout(function() { onFulfilled(3); }, 0); } };
      var nonPromise = 4;

      RSVP.all([promise, syncThenable, asyncThenable, nonPromise]).then(function(results) {
        assert(objectEquals(results, [1, 2, 3, 4]));
        done();
      });
    });

    specify('cancel the array of promise as soon as cancelled', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      var all_promise = RSVP.all(["nonPromiseValue", first, second]);

      all_promise.cancel();

      setTimeout(function() {
        assert(first.isRejected);
        assert(first.rejectedReason instanceof RSVP.CancellationError);
        assert(second.isRejected);
        assert(second.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('cancel the array of promise as soon as rejected', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject("Foo");
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      var all_promise = RSVP.all([first, second]);

      setTimeout(function() {
        assert(first.isRejected);
        assert.equal(first.rejectedReason, "Foo");
        assert(second.isRejected);
        assert(second.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('cancel with non cancellable thenables', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      var all_promise = RSVP.all([{"then": function () {
        return;
      }}, first, second]);

      all_promise.cancel();

      setTimeout(function() {
        assert(first.isRejected);
        assert(first.rejectedReason instanceof RSVP.CancellationError);
        assert(second.isRejected);
        assert(second.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('notifies promises progress', function(done) {
      var firstResolver, secondResolver, count = 0, expected = {};

      var first = new RSVP.Promise(function(resolve, reject, notify) {
        firstResolver = { resolve: resolve, reject: reject, notify: notify };
      });

      var second = new RSVP.Promise(function(resolve, reject, notify) {
        secondResolver = { resolve: resolve, reject: reject, notify: notify };
      });

      expected.index = [0, 1, 1];
      expected.value = [1, 2, 3];

      setTimeout(function() {
        firstResolver.notify(1);
        secondResolver.notify(2);
      }, 0);

      setTimeout(function() {
        secondResolver.notify(3);
      }, 30);

      var all_promise = RSVP.all([first, second]).
        then(null, null, function (value) {
          count += 1;
          assert.equal(value.index, expected.index.shift());
          assert.equal(value.value, expected.value.shift());
        });

      all_promise.cancel();

      setTimeout(function() {
        assert.equal(count, 3);
        done();
      }, 50);
    });
  });

  describe("RSVP.any", function() {
    it('should exist', function() {
      assert(RSVP.any);
    });

    it('throws when not passed an array', function() {
      assert.throws(function () {
        var any = RSVP.any();
      }, TypeError);

      assert.throws(function () {
        var any = RSVP.any('');
      }, TypeError);

      assert.throws(function () {
        var any = RSVP.any({});
      }, TypeError);
    });

    specify('fulfilled only after one of the other promises are fulfilled', function(done) {
      var firstResolved, secondResolved, firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve) {
        firstResolver = resolve;
      });
      first.then(function() {
        firstResolved = true;
      });

      var second = new RSVP.Promise(function(resolve) {
        secondResolver = resolve;
      });
      second.then(function() {
        secondResolved = true;
      });

      setTimeout(function() {
        firstResolver(true);
      }, 20);

      setTimeout(function() {
        secondResolver(true);
      }, 0);

      RSVP.any([first, second]).then(function() {
        assert(firstResolved === undefined);
        assert(secondResolved);
        done();
      });
    });

    specify('rejected as soon as a promise is rejected', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      RSVP.any([first, second]).then(function() {
        assert(false);
      }, function() {
        assert(first.isRejected);
        assert(!second.isResolved);
        done();
      });
    });

    specify('passes the resolved value of the fulfilled promise to the callback', function(done) {
      var firstResolver, secondResolver, thirdResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      var third = new RSVP.Promise(function(resolve, reject) {
        thirdResolver = { resolve: resolve, reject: reject };
      });

      thirdResolver.resolve(3);
      firstResolver.resolve(1);
      secondResolver.resolve(2);

      RSVP.any([first, second, third]).then(function(result) {
        assert.equal(result, 3);
        done();
      });
    });

    specify('resolves an empty array passed to RSVP.any()', function(done) {
      RSVP.any([]).then(function(result) {
        assert.equal(result, undefined);
        done();
      });
    });

    specify('works with a mix of promises and thenables and non-promises', function(done) {
      var promise = new RSVP.Promise(function(resolve) { resolve(1); });
      var syncThenable = { then: function (onFulfilled) { onFulfilled(2); } };
      var asyncThenable = { then: function (onFulfilled) { setTimeout(function() { onFulfilled(3); }, 0); } };
      var nonPromise = 4;

      RSVP.any([promise, syncThenable, asyncThenable, nonPromise]).then(function(result) {
        assert.equal(result, 2);
        done();
      });
    });

    specify('cancel the remaining promises as soon as resolved', function(done) {
      var promise = new RSVP.Promise(function(resolve) { resolve(1); });
      var asyncPromise = new RSVP.Promise(function(resolve) {
        setTimeout(function() { resolve(2); }, 0); });

      RSVP.any([promise, asyncPromise]).then(function(result) {
        assert.equal(result, 1);
      });

      setTimeout(function() {
        assert(asyncPromise.isRejected);
        assert(asyncPromise.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('cancel the array of promise as soon as cancelled', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      var all_promise = RSVP.any([first, second]);

      all_promise.cancel();

      setTimeout(function() {
        assert(first.isRejected);
        assert(first.rejectedReason instanceof RSVP.CancellationError);
        assert(second.isRejected);
        assert(second.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('cancel the array of promise as soon as rejected', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject("Foo");
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      var all_promise = RSVP.any([first, second]);

      setTimeout(function() {
        assert(first.isRejected);
        assert.equal(first.rejectedReason, "Foo");
        assert(second.isRejected);
        assert(second.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('cancel with non cancellable thenables', function(done) {
      var firstResolver, secondResolver;

      var first = new RSVP.Promise(function(resolve, reject) {
        firstResolver = { resolve: resolve, reject: reject };
      });

      var second = new RSVP.Promise(function(resolve, reject) {
        secondResolver = { resolve: resolve, reject: reject };
      });

      setTimeout(function() {
        firstResolver.reject({});
      }, 0);

      setTimeout(function() {
        secondResolver.resolve(true);
      }, 5000);

      var all_promise = RSVP.any([{"then": function () {
        return;
      }}, first, second]);

      all_promise.cancel();

      setTimeout(function() {
        assert(first.isRejected);
        assert(first.rejectedReason instanceof RSVP.CancellationError);
        assert(second.isRejected);
        assert(second.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    specify('notifies promises progress', function(done) {
      var firstResolver, secondResolver, count = 0, expected = {};

      var first = new RSVP.Promise(function(resolve, reject, notify) {
        firstResolver = { resolve: resolve, reject: reject, notify: notify };
      });

      var second = new RSVP.Promise(function(resolve, reject, notify) {
        secondResolver = { resolve: resolve, reject: reject, notify: notify };
      });

      expected.index = [0, 1, 1];
      expected.value = [1, 2, 3];

      setTimeout(function() {
        firstResolver.notify(1);
        secondResolver.notify(2);
      }, 0);

      setTimeout(function() {
        secondResolver.notify(3);
      }, 30);

      var all_promise = RSVP.any([first, second]).
        then(null, null, function (value) {
          count += 1;
          assert.equal(value.index, expected.index.shift());
          assert.equal(value.value, expected.value.shift());
        });

      all_promise.cancel();

      setTimeout(function() {
        assert.equal(count, 3);
        done();
      }, 50);
    });
  });

  describe("RSVP.reject", function(){
    specify("it should exist", function(){
      assert(RSVP.reject);
    });

    describe('it rejects', function(){
      var reason = 'the reason',
      promise = RSVP.reject(reason);

      promise.then(function(){
        assert(false, 'should not fulfill');
      }, function(actualReason){
        assert.equal(reason, actualReason);
      });
    });
  });

  describe("RSVP.onerror", function(){
    var onerror;

    after(function() {
      RSVP.configure('onerror', null);
    });

    it("When provided, any unhandled exceptions are sent to it", function(done) {
      var thrownError = new Error();

      RSVP.configure('onerror', function(error) {
        assert.equal(error, thrownError, "The thrown error is passed in");
        done();
      });

      new RSVP.Promise(function(resolve, reject) {
        reject(thrownError);
      }).then(function() {
        // doesn't get here
      });
    });

    it("When provided, handled exceptions are not sent to it", function(done) {
      var thrownError = new Error();

      RSVP.configure('onerror', function(error) {
        assert(false, "Should not get here");
      });

      new RSVP.Promise(function(resolve, reject) {
        reject(thrownError);
      }).then(null, function(error) {
        assert.equal(error, thrownError, "The handler should handle the error");
        done();
      });
    });
  });

  describe("RSVP.rethrow", function() {
    var onerror;

    after(function() {
      global.setTimeout = oldSetTimeout;
    });

    it("should exist", function() {
      assert(RSVP.rethrow);
    });

    it("rethrows an error", function(done) {
      var thrownError = new Error('I am an error.');

      function expectRejection(reason) {
        done();
        assertEqual(reason, thrownError);
      }

      function doNotExpectFulfillment(value) {
        done();
        assert(false, value);
      }

      global.setTimeout = newSetTimeout;

      RSVP.reject(thrownError).
        fail(RSVP.rethrow).
        then(doNotExpectFulfillment, expectRejection);
    });
  });

  describe("RSVP.resolve", function(){
    specify("it should exist", function(){
      assert(RSVP.resolve);
    });

    describe("1. If x is a promise, adopt its state ", function(){
      specify("1.1 If x is pending, promise must remain pending until x is fulfilled or rejected.", function(done){
        var expectedValue, resolver, thenable, wrapped;

        expectedValue = 'the value';
        thenable = {
          then: function(resolve, reject){
            resolver = resolve;
          }
        };

        wrapped = RSVP.resolve(thenable);

        wrapped.then(function(value){
          assert(value === expectedValue);
          done();
        })
        resolver(expectedValue);

      });

      specify("1.2 If/when x is fulfilled, fulfill promise with the same value.", function(done){
        var expectedValue, thenable, wrapped;

        expectedValue = 'the value';
        thenable = {
          then: function(resolve, reject){
            resolve(expectedValue);
          }
        };

        wrapped = RSVP.resolve(thenable);

        wrapped.then(function(value){
          assert(value === expectedValue);
          done();
        })
      });

      specify("1.3 If/when x is rejected, reject promise with the same reason.", function(done){
        var expectedError, thenable, wrapped;

        expectedError =  new Error();
        thenable = {
          then: function(resolve, reject){
            reject(expectedError);
          }
        };

        wrapped = RSVP.resolve(thenable);

        wrapped.then(null, function(error){
          assert(error === expectedError);
          done();
        });
      });
    });

    describe("2. Otherwise, if x is an object or function,", function(){
      specify("2.1 Let then x.then", function(done){
        var accessCount, resolver, wrapped, thenable;

        accessCount = 0;
        thenable = { };

        // we likely don't need to test this, if the browser doesn't support it
        if (typeof Object.defineProperty !== "function") { done(); return; }

        Object.defineProperty(thenable, 'then', {
          get: function(){
            accessCount++;

            if (accessCount > 1) {
              throw new Error();
            }

            return function(){ };
          }
        });

        assert(accessCount === 0);

        wrapped = RSVP.resolve(thenable);

        setTimeout(function() {
          assert(accessCount === 1);
          done();
        }, 20);
      });

      specify("2.2 If retrieving the property x.then results in a thrown exception e, reject promise with e as the reason.", function(done){
        var wrapped, thenable, expectedError;

        expectedError = new Error();
        thenable = { };

        // we likely don't need to test this, if the browser doesn't support it
        if (typeof Object.defineProperty !== "function") { done(); return; }

        Object.defineProperty(thenable, 'then', {
          get: function(){
            throw expectedError;
          }
        });

        wrapped = RSVP.resolve(thenable);

        wrapped.then(null, function(error){
          assert(error === expectedError, 'incorrect exception was thrown');
          done();
        });
      });

      describe('2.3. If then is a function, call it with x as this, first argument resolvePromise, and second argument rejectPromise, where', function(){
        specify('2.3.1 If/when resolvePromise is called with a value y, run Resolve(promise, y)', function(done){
          var expectedSuccess, resolver, rejector, thenable, wrapped, calledThis;

          thenable = {
            then: function(resolve, reject){
              calledThis = this;
              resolver = resolve;
              rejector = reject;
            }
          };

          expectedSuccess = 'success';
          wrapped = RSVP.resolve(thenable);

          wrapped.then(function(success){
            assert(calledThis === thenable, 'this must be the thenable');
            assert(success === expectedSuccess, 'rejected promise with x');
            done();
          });

          resolver(expectedSuccess);
        });

        specify('2.3.2 If/when rejectPromise is called with a reason r, reject promise with r.', function(done){
          var expectedError, resolver, rejector, thenable, wrapped, calledThis,

          thenable = {
            then: function(resolve, reject){
              calledThis = this;
              resolver = resolve;
              rejector = reject;
            }
          };

          expectedError = new Error();

          wrapped = RSVP.resolve(thenable);

          wrapped.then(null, function(error){
            assert(error === expectedError, 'rejected promise with x');
            done();
          });

          rejector(expectedError);
        });

        specify("2.3.3 If both resolvePromise and rejectPromise are called, or multiple calls to the same argument are made, the first call takes precedence, and any further calls are ignored", function(done){
          var expectedError, expectedSuccess, resolver, rejector, thenable, wrapped, calledThis,
          calledRejected, calledResolved;

          calledRejected = 0;
          calledResolved = 0;

          thenable = {
            then: function(resolve, reject){
              calledThis = this;
              resolver = resolve;
              rejector = reject;
            }
          };

          expectedError = new Error();

          wrapped = RSVP.resolve(thenable);

          wrapped.then(function(){
            calledResolved++;
          }, function(error){
            calledRejected++;
            assert(calledResolved === 0, 'never resolved');
            assert(calledRejected === 1, 'rejected only once');
            assert(error === expectedError, 'rejected promise with x');
          });

          rejector(expectedError);
          rejector(expectedError);

          rejector('foo');

          resolver('bar');
          resolver('baz');

          setTimeout(function(){
            assert(calledRejected === 1, 'only rejected once');
            assert(calledResolved === 0, 'never resolved');
            done();
          }, 50);
        });

        describe("2.3.4 If calling then throws an exception e", function(){
          specify("2.3.4.1 If resolvePromise or rejectPromise have been called, ignore it.", function(done){
            var expectedSuccess, resolver, rejector, thenable, wrapped, calledThis,
            calledRejected, calledResolved;

            expectedSuccess = 'success';

            thenable = {
              then: function(resolve, reject){
                resolve(expectedSuccess);
                throw expectedError;
              }
            };

            wrapped = RSVP.resolve(thenable);

            wrapped.then(function(success){
              assert(success === expectedSuccess, 'resolved not errored');
              done();
            });
          });

          specify("2.3.4.2 Otherwise, reject promise with e as the reason.", function(done) {
            var expectedError, resolver, rejector, thenable, wrapped, calledThis, callCount;

            expectedError = new Error();
            callCount = 0;

            thenable = { then: function() { throw expectedError; } };

            wrapped = RSVP.resolve(thenable);

            wrapped.then(null, function(error){
              callCount++;
              assert(expectedError === error, 'expected the correct error to be rejected');
              done();
            });

            assert(callCount === 0, 'expected async, was sync');
          });
        });
      });

      specify("2.4 If then is not a function, fulfill promise with x", function(done){
        var expectedError, resolver, rejector, thenable, wrapped, calledThis, callCount;

        thenable = { then: 3 };
        callCount = 0;
        wrapped = RSVP.resolve(thenable);

        wrapped.then(function(success){
          callCount++;
          assert(thenable === success, 'fulfilled promise with x');
          done();
        });

        assert(callCount === 0, 'expected async, was sync');
      });
    });

    describe("3. If x is not an object or function, ", function(){
      specify("fulfill promise with x.", function(done){
        var thenable, callCount, wrapped;

        thenable = null;
        callCount = 0;
        wrapped = RSVP.resolve(thenable);

        wrapped.then(function(success){
          callCount++;
          assert(success === thenable, 'fulfilled promise with x');
          done();
        }, function(a){
          assert(false, 'should not also reject');
        });

        assert(callCount === 0, 'expected async, was sync');
      });
    });

    describe("4. If x is a promise, ", function(){
      specify("cancel will also cancel x.", function(done){
        var thenable, wrapped;

        thenable = RSVP.delay();

        wrapped = RSVP.resolve(thenable);

        thenable.fail(function(error){
          assert(error instanceof RSVP.CancellationError, 'expected CancellationError');
          done();
        });

        wrapped.cancel();
      });
    });
  });
});


// thanks to @wizardwerdna for the test case -> https://github.com/tildeio/rsvp.js/issues/66
// Only run these tests in node (phantomjs cannot handle them)
if (typeof module !== 'undefined' && module.exports) {

  describe("using reduce to sum integers using promises", function(){
    var resolve = RSVP.resolve;

    it("should build the promise pipeline without error", function(){
      var array, iters, pZero, i;

      array = [];
      iters = 1000;

      for (i=1; i<=iters; i++) {
        array.push(i);
      }

      pZero = resolve(0);

      array.reduce(function(promise, nextVal) {
        return promise.then(function(currentVal) {
          return resolve(currentVal + nextVal);
        });
      }, pZero);
    });

    it("should get correct answer without blowing the nextTick stack", function(done){
      var pZero, array, iters, result, i;

      pZero = resolve(0);

      array = [];
      iters = 1000;

      for (i=1; i<=iters; i++) {
        array.push(i);
      }

      result = array.reduce(function(promise, nextVal) {
        return promise.then(function(currentVal) {
          return resolve(currentVal + nextVal);
        });
      }, pZero);

      result.then(function(value){
        assert.equal(value, (iters*(iters+1)/2));
        done();
      });
    });
  });

}

describe("`CancellationError`", function () {
  describe("`CancellationError` constructor", function () {
    it('should exist and have length 1', function () {
      assert(RSVP.CancellationError);
      assert.equal(RSVP.CancellationError.length, 1);
    });

    it('should be a constructor', function () {
      var error = new RSVP.CancellationError();
      assert.equal(
        Object.getPrototypeOf(error),
        RSVP.CancellationError.prototype,
        '[[Prototype]] equals CancellationError.prototype'
      );
      assert.equal(
        error.constructor,
        RSVP.CancellationError,
        'constructor property of instances is set correctly'
      );
      assert.equal(
        RSVP.CancellationError.prototype.constructor,
        RSVP.CancellationError,
        'constructor property of prototype is set correctly'
      );
    });

    it('should throw a `TypeError` if not given a string', function () {
      assert.throws(function () {
        var error = new RSVP.CancellationError({});
        // ... jslint forces to use the variable
        error.foo = "bar";
      }, TypeError);
    });

    it('should be throwable', function () {
      assert.throws(function () {
        throw new RSVP.CancellationError();
      }, RSVP.CancellationError);
    });

  });

  describe("`CancellationError` instance", function () {
    it('must be an instance of Error', function () {
      var error = new RSVP.CancellationError();
      assert(error instanceof Error);
    });

    it('must have a `name` property with value `cancel`', function () {
      var error = new RSVP.CancellationError();
      assert.equal(error.name, "cancel");
    });

    it('must have a `message` property with the message value', function () {
      var error = new RSVP.CancellationError("Foo");
      assert.equal(error.message, "Foo");
    });

    it('must have a default `message` property to `Default Message`',
       function () {
        var error = new RSVP.CancellationError();
        assert.equal(error.message, "Default Message");
      });
  });

});

describe("`cancel` on directly created promise", function () {
  it('should be a function', function () {
    var promise = new RSVP.Promise(function (done, fail) {});
    assert.equal(typeof promise.cancel, "function");
  });

  it('should not call `onCancelled` with a `fulfilled` promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve) {
        resolve();
      }, function () {
        cancel_called = true;
      });

    setTimeout(function() {
      promise.cancel();
      setTimeout(function() {
        assert.equal(cancel_called, false);
        done();
      }, 20);
    }, 20);
  });

  it('should not call `onCancelled` with a `rejected` promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve, reject) {
        reject();
      }, function () {
        cancel_called = true;
      });

    setTimeout(function() {
      promise.cancel();
      setTimeout(function() {
        assert.equal(cancel_called, false);
        done();
      }, 20);
    }, 20);
  });

  describe("call on a `pending` promise", function () {
    it('should call `onCancelled` without argument', function (done) {
      var cancel_called = false,
        promise = new RSVP.Promise(function () {
        }, function () {
          cancel_called = true;
          assert.equal(arguments.length, 0);
        });

      setTimeout(function() {
        promise.cancel("Foo");
        setTimeout(function() {
          assert.equal(cancel_called, true);
          done();
        }, 20);
      }, 20);
    });

    it('should reject on canceller exception', function (done) {
      var promise = RSVP.Promise(function() {}, function () {
        throw 'bar error'
      }),
        error;
      promise.then(null, function(e) {
        error = e;
      });
      promise.cancel();
      setTimeout(function() {
        assert.equal(error, 'bar error');
        done();
      }, 20);
    });

    it('should reject with a CancellationError', function (done) {
      var promise = RSVP.Promise(function() {}),
        error;
      promise.then(null, function(e) {
        error = e;
      });
      promise.cancel();
      setTimeout(function() {
        assert(error instanceof RSVP.CancellationError);
        done();
      }, 20);
    });
  });

});

describe("`cancel` on promise created by then", function () {
  it('should be a function', function () {
    var promise = new RSVP.Promise(function (done, fail) {}).then();
    assert.equal(typeof promise.cancel, "function");
  });

  it('should stop parent fulFillment propagation', function (done) {
    var promise = new RSVP.resolve(1),
      promise2 = promise.then();
    promise2.cancel("Foo");

    setTimeout(function() {
      assert.equal(promise2.isRejected, true);
      assert.equal(promise2.isFulfilled, undefined);
      assert(promise2.rejectedReason instanceof RSVP.CancellationError);
      done();
    }, 20);
  });

  it('should stop parent rejection propagation', function (done) {
    var reject_count = 0,
      promise = new RSVP.reject(1),
      promise2 = promise.fail(function () {reject_count += 1;});
    promise2.cancel("Foo");

    setTimeout(function() {
      assert.equal(promise2.isRejected, true);
      assert.equal(promise2.isFulfilled, undefined);
      assert(promise2.rejectedReason instanceof RSVP.CancellationError);
      assert.equal(reject_count, 1);
      done();
    }, 20);
  });

  it('should not be callable twice', function (done) {
    var reject_count = 0,
      promise = new RSVP.reject(1),
      promise2 = promise.fail(function () {reject_count += 1;});
    promise2.cancel("Foo");
    promise2.cancel("Bar");

    setTimeout(function() {
      assert.equal(promise2.isRejected, true);
      assert.equal(promise2.isFulfilled, undefined);
      assert(promise2.rejectedReason instanceof RSVP.CancellationError);
      assert.equal(reject_count, 1);
      done();
    }, 20);
  });

  it('should cancel the pending fulfilled promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve) {resolve();}),
      promise2 = promise.then(function () {
        return RSVP.Promise(function() {}, function () {
          cancel_called = true;
        });
      });

    setTimeout(function() {
      promise2.cancel("Foo");
      setTimeout(function() {
        assert.equal(cancel_called, true);
        assert.equal(promise2.isFulfilled, undefined);
        assert.equal(promise2.isRejected, true);
        assert(promise2.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    }, 20);
  });

  it('should cancel the pending rejected promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve, reject) {reject();}),
      promise2 = promise.then(undefined, function () {
        return RSVP.Promise(function() {}, function () {
          cancel_called = true;
        });
      });

    setTimeout(function() {
      promise2.cancel("Foo");
      setTimeout(function() {
        assert.equal(cancel_called, true);
        assert.equal(promise2.isFulfilled, undefined);
        assert.equal(promise2.isRejected, true);
        assert(promise2.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    }, 20);
  });

});

describe("`cancel` on thenable created by then", function () {
  it('should cancel the pending fulfilled promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve) {resolve();}),
      promise2 = promise.then(function () {
        return {"then": function() {}, "cancel": function () {
          cancel_called = true;
        }};
      });

    setTimeout(function() {
      promise2.cancel("Foo");
      setTimeout(function() {
        assert.equal(cancel_called, true);
        assert.equal(promise2.isFulfilled, undefined);
        assert.equal(promise2.isRejected, true);
        assert(promise2.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    }, 20);
  });

  it('should cancel the pending rejected promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve, reject) {reject();}),
      promise2 = promise.then(undefined, function () {
        return {"then": function() {}, "cancel": function () {
          cancel_called = true;
        }};
      });

    setTimeout(function() {
      promise2.cancel("Foo");
      setTimeout(function() {
        assert.equal(cancel_called, true);
        assert.equal(promise2.isFulfilled, undefined);
        assert.equal(promise2.isRejected, true);
        assert(promise2.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    }, 20);
  });
});

describe("`cancel` on non cancellable thenable created by then", function () {
  it('should cancel the pending fulfilled promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve) {resolve();}),
      promise2 = promise.then(function () {
        return {"then": function() {}};
      });

    setTimeout(function() {
      promise2.cancel("Foo");
      setTimeout(function() {
        assert.equal(promise2.isFulfilled, undefined);
        assert.equal(promise2.isRejected, true);
        assert(promise2.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    }, 20);
  });

  it('should cancel the pending rejected promise', function (done) {
    var cancel_called = false,
      promise = new RSVP.Promise(function (resolve, reject) {reject();}),
      promise2 = promise.then(undefined, function () {
        return {"then": function() {}};
      });

    setTimeout(function() {
      promise2.cancel("Foo");
      setTimeout(function() {
        assert.equal(promise2.isFulfilled, undefined);
        assert.equal(promise2.isRejected, true);
        assert(promise2.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    }, 20);
  });
});

describe("`always` on a promise", function () {
  it('should be a function', function () {
    var promise = new RSVP.Promise(function (done, fail) {});
    assert.equal(typeof promise.always, "function");
  });

  it('should return a new promise', function () {
    var promise = new RSVP.Promise(function (done, fail) {}),
      promise2 = promise.always();
    assert(promise2 instanceof RSVP.Promise);
  });

  it('should fulfill the new promise if the parent is fulfilled', function (done) {
    var promise = new RSVP.Promise(function (accept, fail) {accept("bar")}),
      promise2 = promise.always();

    setTimeout(function() {
      assert.equal(promise2.isFulfilled, true);
      assert.equal(promise2.isRejected, undefined);
      assert.equal(promise2.fulfillmentValue, "bar");
      done();
    }, 20);
  });

  it('should call the function parameter if the parent is fulfilled',
     function (done) {
    var promise = new RSVP.Promise(function (accept, fail) {accept("bar")}),
      parameter_called = false,
      promise2 = promise.always(function () {parameter_called = true});

    setTimeout(function() {
      assert.equal(parameter_called, true);
      done();
    }, 20);
  });

  it('should reject the new promise if the parent is rejected', function (done) {
    var promise = new RSVP.Promise(function (accept, fail) {fail("foo")}),
      promise2 = promise.always();

    setTimeout(function() {
      assert.equal(promise2.isFulfilled, undefined);
      assert.equal(promise2.isRejected, true);
      assert.equal(promise2.rejectedReason, "foo");
      done();
    }, 20);
  });

  it('should call the function parameter if the parent is rejected',
     function (done) {
    var promise = new RSVP.Promise(function (accept, fail) {fail("foo")}),
      parameter_called = false,
      promise2 = promise.always(function () {parameter_called = true});

    setTimeout(function() {
      assert.equal(parameter_called, true);
      done();
    }, 20);
  });
});

describe("`RSVP.delay`", function () {
  it('should be a function', function () {
    assert.equal(typeof RSVP.delay, "function");
  });

  it('should return a new promise', function () {
    var promise = RSVP.delay();
    assert(promise instanceof RSVP.Promise);
  });

  it('should fulfill the new promise after some milliseconds', function (done) {
    var promise = RSVP.delay(30);

    setTimeout(function() {
      assert.equal(promise.isFulfilled, undefined);
      assert.equal(promise.isRejected, undefined);
      setTimeout(function() {
        assert.equal(promise.isFulfilled, true);
        assert.equal(promise.isRejected, undefined);
        done();
      }, 20);
    }, 20);
  });

  it('should allow to cancel the setTimeout', function (done) {
    var promise = RSVP.delay();
    promise.cancel();
    setTimeout(function() {
      // XXX How to check that clearTimeout is called?
      assert.equal(promise.isRejected, true);
      done();
    }, 20);
  });
});

describe("`promise.delay`", function () {
  it('should be a function', function () {
    var promise = new RSVP.Promise(function (accept, fail) {});
    assert.equal(typeof promise.delay, "function");
  });

  it('should return a new promise', function () {
    var promise = new RSVP.Promise(function (accept, fail) {}),
      promise2 = promise.delay();
    assert(promise2 instanceof RSVP.Promise);
  });

  it('should fulfill the new promise after some milliseconds', function (done) {
    var promise = new RSVP.Promise(function (accept, fail) {accept();}),
      promise2 = promise.delay(30);

    setTimeout(function() {
      assert.equal(promise2.isFulfilled, undefined);
      assert.equal(promise2.isRejected, undefined);
      setTimeout(function() {
        assert.equal(promise2.isFulfilled, true);
        assert.equal(promise2.isRejected, undefined);
        done();
      }, 20);
    }, 20);
  });

  it('should propagate the fulfillmentValue', function (done) {
    var promise = new RSVP.Promise(function (accept, fail) {accept("foo");}),
      promise2 = promise.delay(30);

    setTimeout(function() {
      assert.equal(promise2.isFulfilled, true);
      assert.equal(promise2.isRejected, undefined);
      assert.equal(promise2.fulfillmentValue, "foo");
      done();
    }, 40);
  });
});

describe("`RSVP.timeout`", function () {
  it('should be a function', function () {
    assert.equal(typeof RSVP.timeout, "function");
  });

  it('should return a new promise', function () {
    var promise = RSVP.timeout();
    assert(promise instanceof RSVP.Promise);
  });

  it('should reject the new promise after some milliseconds', function (done) {
    var promise = RSVP.timeout(30);

    setTimeout(function() {
      assert.equal(promise.isFulfilled, undefined);
      assert.equal(promise.isRejected, undefined);
      setTimeout(function() {
        assert.equal(promise.isFulfilled, undefined);
        assert.equal(promise.isRejected, true);
        assert.equal(promise.rejectedReason, "Timed out after 30 ms");
        done();
      }, 20);
    }, 20);
  });

  it('should allow to cancel the setTimeout', function (done) {
    var promise = RSVP.timeout();
    promise.cancel();
    setTimeout(function() {
      // XXX How to check that clearTimeout is called?
      assert.equal(promise.isRejected, true);
      done();
    }, 20);
  });
});

describe("`RSVP.Queue`", function () {

  describe("`ResolvedQueueError`", function () {
    describe("`ResolvedQueueError` constructor", function () {
      it('should exist and have length 1', function () {
        assert(RSVP.ResolvedQueueError);
        assert.equal(RSVP.ResolvedQueueError.length, 1);
      });

      it('should be a constructor', function () {
        var error = new RSVP.ResolvedQueueError();
        assert.equal(
          Object.getPrototypeOf(error),
          RSVP.ResolvedQueueError.prototype,
          '[[Prototype]] equals ResolvedQueueError.prototype'
        );
        assert.equal(
          error.constructor,
          RSVP.ResolvedQueueError,
          'constructor property of instances is set correctly'
        );
        assert.equal(
          RSVP.ResolvedQueueError.prototype.constructor,
          RSVP.ResolvedQueueError,
          'constructor property of prototype is set correctly'
        );
      });

      it('should throw a `TypeError` if not given a string', function () {
        assert.throws(function () {
          var error = new RSVP.ResolvedQueueError({});
          // ... jslint forces to use the variable
          error.foo = "bar";
        }, TypeError);
      });

      it('should be throwable', function () {
        assert.throws(function () {
          throw new RSVP.ResolvedQueueError();
        }, RSVP.ResolvedQueueError);
      });

    });

    describe("`ResolvedQueueError` instance", function () {
      it('must be an instance of Error', function () {
        var error = new RSVP.ResolvedQueueError();
        assert(error instanceof Error);
      });

      it('must have a `name` property with value `resolved`', function () {
        var error = new RSVP.ResolvedQueueError();
        assert.equal(error.name, "resolved");
      });

      it('must have a `message` property with the message value', function () {
        var error = new RSVP.ResolvedQueueError("Foo");
        assert.equal(error.message, "Foo");
      });

      it('must have a default `message` property to `Default Message`',
         function () {
          var error = new RSVP.ResolvedQueueError();
          assert.equal(error.message, "Default Message");
        });
    });
  });

  describe("`Queue` constructor", function () {
    it('should exist and have length 0', function () {
      assert(RSVP.Queue);
      assert.equal(RSVP.Queue.length, 0);
    });

    it('should be a constructor', function () {
      var queue = new RSVP.Queue();
      assert.equal(
        Object.getPrototypeOf(queue),
        RSVP.Queue.prototype,
        '[[Prototype]] equals Queue.prototype'
      );
      assert.equal(
        queue.constructor,
        RSVP.Queue,
        'constructor property of instances is set correctly'
      );
      assert.equal(
        RSVP.Queue.prototype.constructor,
        RSVP.Queue,
        'constructor property of prototype is set correctly'
      );
    });

    it('should return a new promise', function () {
      var promise = new RSVP.Queue();
      assert(promise instanceof RSVP.Queue);
      assert(promise instanceof RSVP.Promise);
    });

    it('should work without `new`', function() {
      var promise = RSVP.Queue()
        .push(function() { return 'value'; });

      promise.then(function(value) {
        assert.equal(value, 'value');
      });
    });
  });

  describe("`push`", function () {
    it('should be a function', function () {
      var queue = new RSVP.Queue();
      assert.equal(typeof queue.push, "function");
    });

    it('should return the same queue instance', function () {
      var queue = new RSVP.Queue(),
        result = queue.push();
      assert.equal(result, queue);
    });

    it('should throw if the queue is fulfilled', function (done) {
      var queue = new RSVP.Queue();

      setTimeout(function() {
        assert.throws(function () {
          queue.push();
        }, RSVP.ResolvedQueueError);
        done();
      }, 20);
    });

    it('should throw if the queue is rejected', function (done) {
      var queue = new RSVP.Queue();
      queue.cancel();

      setTimeout(function() {
        assert.throws(function () {
          queue.push();
        }, RSVP.ResolvedQueueError);
        done();
      }, 20);
    });
  });

  describe("`Queue` instance", function () {
    it('should be fulfilled without any push call', function (done) {
      var queue = new RSVP.Queue();

      setTimeout(function() {
        assert.equal(queue.isFulfilled, true);
        done();
      }, 20);
    });

    it('should be fulfilled when the queue is empty', function (done) {
      var queue = new RSVP.Queue();
      queue.push(function () {
        return RSVP.delay(30);
      });

      setTimeout(function() {
        assert.equal(queue.isFulfilled, undefined);
        setTimeout(function() {
          assert.equal(queue.isFulfilled, true);
          done();
        }, 20);
      }, 20);
    });

    it('get the fulfillmentValue of the last queue entry', function (done) {
      var queue = new RSVP.Queue();
      queue.push(function () {
        return "foo"
      });

      setTimeout(function() {
        assert.equal(queue.isFulfilled, true);
        assert.equal(queue.fulfillmentValue, "foo");
        done();
      }, 20);
    });

    it('resolve value is propagated to the next entry', function (done) {
      var queue = new RSVP.Queue(),
        pushed_result;
      queue.push(function () {
        return "foo";
      });
      queue.push(function (value) {
        pushed_result = value;
        return "bar";
      });

      setTimeout(function() {
        assert.equal(pushed_result, "foo");
        assert.equal(queue.isFulfilled, true);
        assert.equal(queue.fulfillmentValue, "bar");
        done();
      }, 20);
    });

    it('should be rejected if one entry is rejected', function (done) {
      var queue = new RSVP.Queue();
      queue.push(function () {
        return RSVP.timeout(30);
      });

      setTimeout(function() {
        assert.equal(queue.isRejected, undefined);
        setTimeout(function() {
          assert.equal(queue.isRejected, true);
          done();
        }, 20);
      }, 20);
    });

    it('get the rejectedReason from the rejected entry', function (done) {
      var queue = new RSVP.Queue();
      queue.push(function () {
        return RSVP.timeout(1);
      });

      setTimeout(function() {
        assert.equal(queue.isRejected, true);
        assert.equal(queue.rejectedReason, "Timed out after 1 ms");
        done();
      }, 20);
    });

    it('reject cancels the remaining promise', function (done) {
      var queue = new RSVP.Queue(),
        pushed_result;
      queue.push(function () {
        return RSVP.timeout(1);
      });
      queue.push(undefined, function (value) {
        pushed_result = value;
        throw value;
      });

      setTimeout(function() {
        assert.equal(pushed_result, "Timed out after 1 ms");
        assert.equal(queue.isRejected, true);
        assert.equal(queue.rejectedReason, "Timed out after 1 ms");
        done();
      }, 20);
    });

    it('cancel also cancels the remaining promise', function (done) {
      var queue = new RSVP.Queue(),
        pushed_result;
      queue.push(function () {
        return RSVP.timeout(1);
      });
      queue.push(undefined, function (value) {
        pushed_result = value;
        throw value;
      });
      queue.cancel();

      setTimeout(function() {
        assert(pushed_result instanceof RSVP.CancellationError);
        assert.equal(queue.isRejected, true);
        assert(queue.rejectedReason instanceof RSVP.CancellationError);
        done();
      }, 20);
    });

    it("notify should be propagated until `then`", function (done) {
      var queue = new RSVP.Queue(), i = 0, responses = [];

      queue.push(function () {
        return new RSVP.Promise(function (resolve, reject, notify) {
          notify("hello");
        });
      });

      queue.push(null, null, function (notification) {
        responses.push(1);
        responses.push(notification);
        return notification;
      });

      queue.push(function () {
        return;
      });

      queue.then(null, null, function (notification) {
        responses.push(2);
        responses.push(notification);
      });

      setTimeout(function () {
        assert.equal(responses[0], 1);
        assert.equal(responses[1], "hello", "Notification");
        assert.equal(responses[2], 2);
        assert.equal(responses[3], "hello", "Notification");
        done();
      }, 50);
    });
  });
});
